#  Copyright (C) 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  cornelius@privacyidea.org
#
#  2019-47-14 Paul Lettich <paul.lettich@netknights.it>
#             Remove hash calculation and switch to passlib
#  2016-07-15 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add sha512 PHP hash as suggested by Rick Romero
#  2016-04-08 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Simplifying out of bounds check
#             Avoid repetition in comparison
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This is the resolver to find users in SQL databases.

The file is tested in tests/test_lib_resolver.py
"""

import logging
import yaml
import binascii
import re

from privacyidea.lib.resolvers.UserIdResolver import UserIdResolver

from sqlalchemy import (Integer, cast, String, MetaData, Table, and_,
                        create_engine, select, insert, delete)
from sqlalchemy.orm import sessionmaker, scoped_session

import traceback
import hashlib
from privacyidea.lib.pooling import get_engine
from privacyidea.lib.lifecycle import register_finalizer
from privacyidea.lib.utils import (is_true, censor_connect_string,
                                   convert_column_to_unicode)
from passlib.context import CryptContext
from passlib.utils import h64
from passlib.utils.compat import uascii_to_str
from passlib.utils.compat import unicode as pl_unicode
from passlib.utils import to_unicode
import passlib.utils.handlers as uh
import passlib.exc as exc
from passlib.registry import register_crypt_handler
from passlib.handlers.ldap_digests import _SaltedBase64DigestHelper


class phpass_drupal(uh.HasRounds, uh.HasSalt, uh.GenericHandler):  # pragma: no cover
    """This class implements the PHPass Portable Hash (Drupal version), and follows the
    :ref:`password-hash-api`.
    """
    name = "phpass_drupal"
    setting_kwds = ("salt", "rounds")
    checksum_chars = uh.HASH64_CHARS
    checksum_size = 43

    min_salt_size = max_salt_size = 8
    salt_chars = uh.HASH64_CHARS

    default_rounds = 19
    min_rounds = 7
    max_rounds = 30
    rounds_cost = "log2"

    ident = '$S$'

    @classmethod
    def from_string(cls, hash):
        hash = to_unicode(hash, "ascii", "hash")
        ident, data = hash[0:3], hash[3:]
        if ident != cls.ident:
            raise exc.InvalidHashError()
        rounds, salt, chk = data[0], data[1:9], data[9:]
        return cls(
            rounds=h64.decode_int6(rounds.encode("ascii")),
            salt=salt,
            checksum=chk or None,
        )

    def to_string(self):
        hash = "%s%s%s%s" % (self.ident,
                              h64.encode_int6(self.rounds).decode("ascii"),
                              self.salt,
                              self.checksum or '')
        return uascii_to_str(hash)

    def _calc_checksum(self, secret):
        if isinstance(secret, pl_unicode):
            secret = secret.encode("utf-8")
        real_rounds = 1 << self.rounds
        result = hashlib.sha512(self.salt.encode("ascii") + secret).digest()
        r = 0
        while r < real_rounds:
            result = hashlib.sha512(result + secret).digest()
            r += 1
        return h64.encode_bytes(result).decode("ascii")[:self.checksum_size]


class ldap_salted_sha256_pi(_SaltedBase64DigestHelper):
    name = 'ldap_salted_sha256_pi'
    ident = '{SSHA256}'
    checksum_size = 32
    # passlib sets the max_salt_size for SSHA256 to 16:
    # <https://foss.heptapod.net/python-libs/passlib/-/blob/branch/stable/passlib/handlers/ldap_digests.py#L71>
    # But we have encountered hashes with longer salt sizes. Since we set this
    # password hashing function when we create an internal db resolver,
    # we should be able to verify these hashes as well.
    max_salt_size = 32
    _hash_func = hashlib.sha256
    _hash_regex = re.compile(r"^\{SSHA256\}(?P<tmp>[+/a-zA-Z0-9]{48,}={0,2})$")


register_crypt_handler(phpass_drupal)
register_crypt_handler(ldap_salted_sha256_pi)


# The list of supported password hash types for verification (passlib handler)
pw_ctx = CryptContext(schemes=['phpass',
                               'phpass_drupal',
                               'ldap_salted_sha1',
                               'ldap_salted_sha256_pi',
                               'ldap_salted_sha512',
                               'ldap_sha1',
                               'md5_crypt',
                               'bcrypt',
                               'sha512_crypt',
                               'sha256_crypt',
                               'hex_sha256',
                               ])

# List of supported password hash types for hash generation (name to passlib handler id)
hash_type_dict = {"PHPASS": 'phpass',
                  "SHA": 'ldap_sha1',
                  "SSHA": 'ldap_salted_sha1',
                  "SSHA256": 'ldap_salted_sha256_pi',
                  "SSHA512": 'ldap_salted_sha512',
                  "OTRS": 'hex_sha256',
                  "SHA256CRYPT": 'sha256_crypt',
                  "SHA512CRYPT": 'sha512_crypt',
                  "MD5CRYPT": 'md5_crypt',
                  }

log = logging.getLogger(__name__)


class IdResolver (UserIdResolver):

    searchFields = {"username": "text",
                    "userid": "numeric",
                    "phone": "text",
                    "mobile": "text",
                    "surname": "text",
                    "givenname": "text",
                    "email": "text",
                    "description": "text",
                    }

    # If the resolver could be configured editable
    updateable = True

    @staticmethod
    def setup(config=None, cache_dir=None):
        """
        this setup hook is triggered, when the server
        starts to serve the first request

        :param config: the privacyidea config
        :type  config: the privacyidea config dict
        """
        log.info("Setting up the SQLResolver")

    def __init__(self):
        self.resolverId = ""
        self.server = ""
        self.driver = ""
        self.database = ""
        self.port = 0
        self.limit = 100
        self.user = ""
        self.password = "" # nosec B105 # default parameter
        self.table = ""
        self.TABLE = None
        self.map = {}
        self.reverse_map = {}
        self.where = ""
        self.encoding = ""
        self.conParams = ""
        self.connect_string = ""
        self.session = None
        self.pool_size = 10
        self.pool_timeout = 120
        self.pool_recycle = 7200
        self.engine = None
        self._editable = False
        self.password_hash_type = None
        return

    def getSearchFields(self):
        return self.searchFields

    @staticmethod
    def _append_where_filter(conditions, table, where):
        """
        Append contents of WHERE statement to the list of filter conditions

        :param conditions: filter conditions
        :type conditions: list
        :return: list of filter conditions
        """
        if where:
            parts = re.split(' and ', where, flags=re.IGNORECASE)
            for part in parts:
                # this might result in errors if the
                # administrator enters nonsense
                (w_column, w_cond, w_value) = part.split()
                if w_cond.lower() == "like":
                    conditions.append(table.columns[w_column].like(w_value))
                elif w_cond == "==":
                    conditions.append(table.columns[w_column] == w_value)
                elif w_cond == ">":
                    conditions.append(table.columns[w_column] > w_value)
                elif w_cond == "<":
                    conditions.append(table.columns[w_column] < w_value)

        return conditions

    def checkPass(self, uid, password):
        """
        This function checks the password for a given uid.

        :param uid: uid of the user for which the password should be checked
        :type uid: str
        :param password: the password to check
        :type password: str
        :return: True if password matches the saved password hash, False otherwise
        :rtype: bool
        """

        res = False
        userinfo = self.getUserInfo(uid)

        database_pw = userinfo.get("password", "XXXXXXX")

        # remove owncloud hash format identifier (currently only version 1)
        database_pw = re.sub(r'^1\|', '', database_pw)

        # translate lower case hash identifier to uppercase
        database_pw = re.sub(r'^{([a-z0-9]+)}',
                             lambda match: '{{{}}}'.format(match.group(1).upper()),
                             database_pw)

        try:
            res = pw_ctx.verify(password, database_pw)
        except ValueError as _e:
            # if the hash could not be identified / verified, just return False
            pass

        return res

    def getUserInfo(self, userId):
        """
        This function returns all user info for a given userid/object.

        :param userId: The userid of the object
        :type userId: string
        :return: A dictionary with the keys defined in self.map
        :rtype: dict
        """
        userinfo = {}

        try:
            conditions = [self._get_userid_filter(userId)]
            conditions = self._append_where_filter(conditions, self.TABLE,
                                                   self.where)
            filter_condition = and_(*conditions)
            result = self.session.execute(select(self.TABLE).filter(filter_condition))

            for r in result.mappings():
                if userinfo:  # pragma: no cover
                    raise Exception("More than one user with userid {0!s} found!".format(userId))
                userinfo = self._get_user_from_mapped_object(r)
        except Exception as exx:  # pragma: no cover
            log.error("Could not get the user information: {0!r}".format(exx))

        return userinfo

    def _get_userid_filter(self, userId):
        column = self.TABLE.columns[self.map.get("userid")]
        if isinstance(column.type, String):
            return column == str(userId)
        elif isinstance(column.type, Integer):
            # since our user ID is usually a string we need to cast
            return column == int(userId)

        # otherwise we cast the column to string (in case of postgres UUIDs)
        return cast(column, String).like(userId)

    def getUsername(self, userId):
        """
        Returns the username/loginname for a given userid

        :param userId: The userid in this resolver
        :type userId: string
        :return: username
        :rtype: string
        """
        info = self.getUserInfo(userId)
        return info.get('username', "")

    def getUserId(self, LoginName):
        """
        resolve the loginname to the userid.

        :param LoginName: The login name from the credentials
        :type LoginName: string
        :return: UserId as found for the LoginName
        :rtype: str
        """
        userid = ""

        try:
            conditions = []
            column = self.map.get("username")
            conditions.append(self.TABLE.columns[column].like(LoginName))
            conditions = self._append_where_filter(conditions, self.TABLE,
                                                   self.where)
            filter_condition = and_(*conditions)
            result = self.session.execute(select(self.TABLE).filter(filter_condition))

            for r in result.mappings():
                if userid != "":    # pragma: no cover
                    raise Exception("More than one user with loginname"
                                    " %s found!" % LoginName)
                user = self._get_user_from_mapped_object(r)
                userid = convert_column_to_unicode(user["userid"])
        except Exception as exx:    # pragma: no cover
            log.error("Could not get the user ID: {0!r}".format(exx))

        return userid

    def _get_user_from_mapped_object(self, ro):
        """
        :param ro: row
        :type ro: Mapped Object
        :return: User
        :rtype: dict
        """
        user = {}
        try:
            if self.map.get("userid") in ro:
                user["id"] = ro[self.map.get("userid")]
        except UnicodeEncodeError:  # pragma: no cover
            log.error("Failed to convert user: {0!r}".format(ro))
            log.debug("{0!s}".format(traceback.format_exc()))

        for key in self.map.keys():
            try:
                raw_value = ro.get(self.map.get(key))
                if raw_value:
                    if key == 'userid':
                        val = convert_column_to_unicode(raw_value)
                    elif isinstance(raw_value, bytes):
                        val = raw_value.decode(self.encoding)
                    else:
                        val = raw_value
                    user[key] = val

            except UnicodeDecodeError:  # pragma: no cover
                user[key] = "decoding_error"
                log.error("Failed to convert user: {0!r}".format(ro))
                log.debug("{0!s}".format(traceback.format_exc()))

        return user

    def getUserList(self, searchDict=None):
        """
        :param searchDict: A dictionary with search parameters
        :type searchDict: dict
        :return: list of users, where each user is a dictionary
        """
        users = []
        conditions = []
        if searchDict is None:
            searchDict = {}
        for key in searchDict.keys():
            column = self.map.get(key)
            value = searchDict.get(key)
            value = value.replace("*", "%")
            conditions.append(self.TABLE.columns[column].like(value))

        conditions = self._append_where_filter(conditions, self.TABLE,
                                               self.where)
        filter_condition = and_(*conditions)

        result = self.session.execute(select(self.TABLE).
                                      filter(filter_condition).
                                      limit(self.limit))

        for r in result.mappings():
            user = self._get_user_from_mapped_object(r)
            if "userid" in user:
                # Remove the "password" attribute
                user.pop("password", None)
                users.append(user)

        return users

    def getResolverId(self):
        """
        Returns the resolver Id
        This should be an Identifier of the resolver, preferable the type
        and the name of the resolver.

        :return: identifier of the resolver
        :rtype: str
        """
        # Take the following parts, join them with the NULL byte and return
        # the hexlified SHA-1 digest
        id_parts = (self.connect_string,
                    str(self.pool_size),
                    str(self.pool_recycle),
                    str(self.pool_timeout))
        id_str = "\x00".join(id_parts)
        resolver_id = binascii.hexlify(hashlib.sha1(id_str.encode('utf8')).digest())  # nosec B324 # hash used as unique identifier
        return "sql." + resolver_id.decode('utf8')

    @staticmethod
    def getResolverClassType():
        return 'sqlresolver'

    @staticmethod
    def getResolverType():
        return IdResolver.getResolverClassType()

    def loadConfig(self, config):
        """
        Load the config from conf.

        :param config: The configuration from the Config Table
        :type config: dict
        """
        self.server = config.get('Server', "")
        self.driver = config.get('Driver', "")
        self.database = config.get('Database', "")
        self.resolverId = self.database
        self.port = config.get('Port', "")
        self.limit = config.get('Limit', 100)
        self.user = config.get('User', "")
        self.password = config.get('Password', "")
        self.table = config.get('Table', "")
        self._editable = config.get("Editable", False)
        self.password_hash_type = config.get("Password_Hash_Type", "SSHA256")
        usermap = config.get('Map', {})
        self.map = yaml.safe_load(usermap)
        self.reverse_map = {v: k for k, v in self.map.items()}
        self.where = config.get('Where', "")
        self.encoding = str(config.get('Encoding') or "latin1")
        self.conParams = config.get('conParams', "")
        self.pool_size = int(config.get('poolSize') or 5)
        self.pool_timeout = int(config.get('poolTimeout') or 10)
        # recycle SQL connections after 2 hours by default
        # (necessary for MySQL servers, which terminate idle connections after some hours)
        self.pool_recycle = int(config.get('poolRecycle') or 7200)

        # create the connect-string like
        params = {'Port': self.port,
                  'Password': self.password,
                  'conParams': self.conParams,
                  'Driver': self.driver,
                  'User': self.user,
                  'Server': self.server,
                  'Database': self.database}
        self.connect_string = self._create_connect_string(params)

        # get an engine from the engine registry, using self.getResolverId() as the key,
        # which involves the connect-string and the pool settings.
        self.engine = get_engine(self.getResolverId(), self._create_engine)
        # We use ``scoped_session``.
        self.session = scoped_session(sessionmaker(bind=self.engine))()
        # Session should be closed on teardown
        register_finalizer(self.session.close)
        self.session._model_changes = {}

        table_parts = self.table.split(".")
        schema = table_parts[0] if len(table_parts) > 1 else None
        self.table = table_parts[-1]
        log.debug("Loading table {0!s} from schema {1!s}".format(self.table, schema))
        self.TABLE = Table(self.table, MetaData(), autoload_with=self.engine, schema=schema)
        return self

    def _create_engine(self):
        log.debug("using the connect string "
                  "{0!s}".format(censor_connect_string(self.connect_string)))
        log.debug("using pool_size={0!s}, pool_timeout={1!s}, pool_recycle={2!s}".format(
            self.pool_size, self.pool_timeout, self.pool_recycle))
        try:
            engine = create_engine(self.connect_string,
                                   encoding=self.encoding,
                                   convert_unicode=False,
                                   pool_size=self.pool_size,
                                   pool_recycle=self.pool_recycle,
                                   pool_timeout=self.pool_timeout)
        except TypeError:
            # The DB Engine/Poolclass might not support the pool_size.
            log.debug("connecting without pool_size.")
            engine = create_engine(self.connect_string,
                                   encoding=self.encoding,
                                   convert_unicode=False)
        return engine

    @classmethod
    def getResolverClassDescriptor(cls):
        descriptor = {}
        typ = cls.getResolverType()
        descriptor['clazz'] = "useridresolver.SQLIdResolver.IdResolver"
        descriptor['config'] = {'Server': 'string',
                                'Driver': 'string',
                                'Database': 'string',
                                'User': 'string',
                                'Password': 'password',
                                'Password_Hash_Type': 'string',
                                'Port': 'int',
                                'Limit': 'int',
                                'Table': 'string',
                                'Map': 'string',
                                'Where': 'string',
                                'Editable': 'int',
                                'poolTimeout': 'int',
                                'poolSize': 'int',
                                'poolRecycle': 'int',
                                'Encoding': 'string',
                                'conParams': 'string'}
        return {typ: descriptor}

    @staticmethod
    def getResolverDescriptor():
        return IdResolver.getResolverClassDescriptor()

    @staticmethod
    def _create_connect_string(param):
        """
        create the connect-string.

        Port, Password, conParams, Driver, User,
        Server, Database
        """
        port = ""
        password = "" # nosec B105 # default parameter
        conParams = ""
        if param.get("Port"):
            port = ":{0!s}".format(param.get("Port"))
        if param.get("Password"):
            password = ":{0!s}".format(param.get("Password"))
        if param.get("conParams"):
            conParams = "?{0!s}".format(param.get("conParams"))
        connect_string = "{0!s}://{1!s}{2!s}{3!s}{4!s}{5!s}/{6!s}{7!s}".format(param.get("Driver") or "",
                                                   param.get("User") or "",
                                                   password,
                                                   "@" if (param.get("User")
                                                           or
                                                           password) else "",
                                                   param.get("Server") or "",
                                                   port,
                                                   param.get("Database") or "",
                                                   conParams)
        return connect_string

    @classmethod
    def testconnection(cls, param):
        """
        This function lets you test the to be saved SQL connection.

        :param param: A dictionary with all necessary parameter
                        to test the connection.
        :type param: dict
        :return: Tuple of success and a description
        :rtype: (bool, string)

        Parameters are: Server, Driver, Database, User, Password, Port,
                        Limit, Table, Map
                        Where, Encoding, conParams

        """
        num = -1
        try:
            connect_string = cls._create_connect_string(param)
            log.info("using the connect string {0!s}".format(censor_connect_string(connect_string)))
            engine = create_engine(connect_string)
            # create a configured "Session" class
            session = scoped_session(sessionmaker(bind=engine))()
        except Exception as e:
            log.warning(f"Unable to connect to database: {e}")
            return -1, "Unable to connect to database."

        table_parts = param.get("Table").split(".")
        schema = table_parts[0] if len(table_parts) > 1 else None
        table_name = table_parts[-1]
        log.debug("Loading table {0!s} from schema {1!s}".format(table_name, schema))

        try:
            TABLE = Table(table_name, MetaData(), autoload_with=engine, schema=schema)
            conditions = cls._append_where_filter([], TABLE,
                                                  param.get("Where"))
            filter_condition = and_(*conditions)
            result = session.query(TABLE).filter(filter_condition).count()

            num = result
            desc = "Found {0:d} users.".format(num)
        except Exception as e:
            log.warning(f"Failed to retrieve users: {e!r}")
            desc = f"Failed to retrieve users."
        finally:
            # We do not want any leftover DB connection, so we first need to close
            # the session such that the DB connection gets returned to the pool (it
            # is still open at that point!) and then dispose the engine such that the
            # checked-in connection gets closed.
            session.close()
            engine.dispose()

        return num, desc

    def add_user(self, attributes=None):
        """
        Add a new user to the SQL database.

        attributes are these
        "username", "surname", "givenname", "email",
        "mobile", "phone", "password"

        :param attributes: Attributes according to the attribute mapping
        :return: The new UID of the user. The UserIdResolver needs to
        determine the way how to create the UID.
        """
        attributes = attributes or {}
        # TODO: add try/except
        kwargs = self.prepare_attributes_for_db(attributes)
        log.info("Insert new user with attributes {0!s}".format(kwargs))
        r = self.session.execute(insert(self.TABLE).values(**kwargs))
        self.session.commit()
        # Return the UID of the new object
        return r.inserted_primary_key[self.map.get("userid")]

    def prepare_attributes_for_db(self, attributes):
        """
        Given a dictionary of attributes, return a dictionary
        mapping columns to values.
        If the attributes contain a password, hash the password according to the
        configured password hash type.

        :param attributes: attributes dictionary
        :return: dictionary with column name as keys
        """
        attributes = attributes.copy()
        if "password" in attributes:
            attributes["password"] = hash_password(attributes["password"],
                                                   self.password_hash_type)
        columns = {}
        for fieldname in attributes:
            if fieldname in self.map:
                columns[self.map[fieldname]] = attributes[fieldname]
        return columns

    def delete_user(self, uid):
        """
        Delete a user from the SQL database.

        The user is referenced by the user id.
        :param uid: The uid of the user object, that should be deleted.
        :type uid: basestring
        :return: Returns True in case of success
        :rtype: bool
        """
        res = True
        try:
            conditions = [self._get_userid_filter(uid)]
            conditions = self._append_where_filter(conditions, self.TABLE,
                                                   self.where)
            filter_condition = and_(*conditions)
            self.session.execute(delete(self.TABLE).where(filter_condition))
            self.session.commit()
            log.info('Deleted user with uid: {0!s}'.format(uid))
        except Exception as exx:
            log.error("Error deleting user: {0!s}".format(exx))
            res = False
        return res

    def update_user(self, uid, attributes=None):
        """
        Update an existing user.
        This function is also used to update the password. Since the
        attribute mapping know, which field contains the password,
        this function can also take care for password changing.

        Attributes that are not contained in the attributes dict are not
        modified.

        :param uid: The uid of the user object in the resolver.
        :type uid: basestring
        :param attributes: Attributes to be updated.
        :type attributes: dict
        :return: True in case of success
        """
        r = False
        attributes = attributes or {}
        try:
            params = self.prepare_attributes_for_db(attributes)
            kwargs = {self.map.get("userid"): uid}
            r = self.session.query(self.TABLE).filter_by(**kwargs).update(params)
            self.session.commit()
            log.info('Updated user attributes for user with uid {0!s}'.format(uid))
        except Exception as exx:
            log.error('Error updating user attributes for user with uid {0!s}: '
                      '{1!s}'.format(uid, exx))
            log.debug('Error updating attributes {0!s}'.format(attributes), exc_info=True)

        return r

    @property
    def editable(self):
        """
        Return true, if the instance of the resolver is configured editable

        :return:
        :rtype: bool
        """
        # Depending on the database this might look different
        # Usually this is "1"
        return is_true(self._editable)


def hash_password(password, hashtype):
    """
    Hash a password with phppass, SHA, SSHA, SSHA256, SSHA512, OTRS

    :param password: The password in plain text
    :type password: str
    :param hashtype: One of the hash types as string
    :type hashtype: str
    :return: The hashed password
    :rtype: str
    """
    hashtype = hashtype.upper()
    try:
        password = pw_ctx.handler(hash_type_dict[hashtype]).hash(password)
    except KeyError as _e:  # pragma: no cover
        raise Exception("Unsupported password hashtype '{0!s}'. "
                        "Use one of {1!s}.".format(hashtype, hash_type_dict.keys()))

    return password



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
import json
import logging

import cryptography.x509
from cryptography.hazmat._oid import NameOID
from cryptography.hazmat.primitives import serialization
from cryptography.x509 import Certificate, NameAttribute
from flask_babel import lazy_gettext
from webauthn import (generate_registration_options,
                      options_to_json, verify_registration_response, verify_authentication_response)
from webauthn.authentication.verify_authentication_response import VerifiedAuthentication
from webauthn.helpers import bytes_to_base64url, base64url_to_bytes, parse_attestation_object
from webauthn.helpers.cose import COSEAlgorithmIdentifier
from webauthn.helpers.exceptions import InvalidRegistrationResponse, InvalidAuthenticationResponse, InvalidJSONStructure
from webauthn.helpers.structs import (AttestationConveyancePreference, AuthenticatorSelectionCriteria,
                                      ResidentKeyRequirement,
                                      PublicKeyCredentialDescriptor, UserVerificationRequirement, AttestationObject,
                                      PublicKeyCredentialCreationOptions)
from webauthn.registration.verify_registration_response import VerifiedRegistration

from privacyidea.api.lib.utils import get_optional, get_required, get_required_one_of, get_optional_one_of
from privacyidea.lib import _
from privacyidea.lib import fido2
from privacyidea.lib.challenge import get_challenges
from privacyidea.lib.config import get_from_config
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib.error import EnrollmentError, ParameterError, ERROR
from privacyidea.lib.fido2.config import FIDO2ConfigOptions
from privacyidea.lib.fido2.policy_action import FIDO2PolicyAction, PasskeyAction
from privacyidea.lib.fido2.token_info import FIDO2TokenInfo
from privacyidea.lib.fido2.util import hash_credential_id, save_credential_id_hash
from privacyidea.lib.log import log_with
from privacyidea.lib.policy import ACTION, SCOPE
from privacyidea.lib.tokenclass import TokenClass, ROLLOUTSTATE, AUTHENTICATIONMODE, CLIENTMODE
from privacyidea.models import Challenge

log = logging.getLogger(__name__)


class PasskeyTokenClass(TokenClass):
    """
    Implements a token class for passkeys (fido2). This is very similar to the webauthn token class, but uses a lib
    for registration and authentication. It is less configurable, always requires resident key and uses excluded
    credentials by default.
    It shares the following policy configuration with the webauthn token class:
        - RP_ID
        - RP_NAME
        - USER_VERIFICATION_REQUIREMENT (default: PREFERRED)
        - PUBLIC_KEY_CREDENTIAL_ALGORITHMS (default: ECDSA_SHA_256, RSASSA_PKCS1_v1_5_SHA_256)
    """

    mode = [AUTHENTICATIONMODE.CHALLENGE]
    client_mode = CLIENTMODE.WEBAUTHN

    def __init__(self, db_token):
        super().__init__(db_token)
        self.set_type(self.get_class_type())

    @staticmethod
    def get_class_type():
        return "passkey"

    @staticmethod
    def get_class_prefix():
        return "PIPK"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        Returns a dict with information about the passkey token class and related policy options.
        The parameter ret can be used to specify the "section" of the information that should be returned.
        """
        res = {
            "type": "passkey",
            "title": "Passkey",
            "description": _("Passkey: A secret stored on a device, unlocked with biometrics."),
            "init": {},
            'config': {},
            'user': ['enroll'],
            'ui_enroll': ["admin", "user"],
            'policy': {
                SCOPE.AUTH: {
                    ACTION.CHALLENGETEXT: {
                        'type': 'str',
                        'desc': _("Alternative challenge message to use when authenticating with a passkey."
                                  "You can also use tags for replacement, "
                                  "check the documentation for more details.")
                    },
                    PasskeyAction.EnableTriggerByPIN: {
                        'type': 'bool',
                        'desc': _("When enabled, passkey token can be triggered with the PIN or via the "
                                  "/validate/triggerchallenge endpoint. For privacyIDEA plugins, "
                                  "this is not recommended. It is advised to use a condition, for example on a "
                                  "user-agent, with this policy."),
                    }
                },
                SCOPE.ENROLL: {
                    PasskeyAction.AttestationConveyancePreference: {
                        'type': 'str',
                        'desc': _("Request attestation from the authenticator during the registration. The attestation "
                                  "certificate will be saved in the token info. The default value is 'none'."),
                        'value': [v for v in AttestationConveyancePreference],
                        'group': 'WebAuthn'
                    }
                }
            }
        }
        return res.get(key, {}) if key else res

    @log_with(log)
    def get_init_detail(self, params=None, user=None):
        """
        First step of enrollment: Returns the registration data for the passkey token.
        Also creates a challenge in the database which has to be verified in the second step.
        The following parameters are required in params:
        - "webauthn_relying_party_id" (FIDO2PolicyAction.RELYING_PARTY_ID)
        - "webauthn_relying_party_name" (FIDO2PolicyAction.RELYING_PARTY_NAME)

        The following parameters are optional in params to customize the registration:
        - "registered_credential_ids": A list of credential IDs that are already registered with the user.
        - FIDO2PolicyAction.PUBLIC_KEY_CREDENTIAL_ALGORITHMS (default: ECDSA_SHA_256, RSASSA_PKCS1_v1_5_SHA_256)
        - FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT (default: PREFERRED)
        - PasskeyAction.AttestationConveyancePreference (default: NONE)
        """
        if self.token.rollout_state == ROLLOUTSTATE.CLIENTWAIT:
            token_user = self.user or user
            if not token_user:
                raise ParameterError("User must be provided for passkey enrollment!",
                                     id=ERROR.PARAMETER_USER_MISSING)

            rp_id = get_required(params, FIDO2PolicyAction.RELYING_PARTY_ID)
            rp_name = get_required(params, FIDO2PolicyAction.RELYING_PARTY_NAME)

            response_detail: dict = TokenClass.get_init_detail(self, params, token_user)
            response_detail['rollout_state'] = self.token.rollout_state
            nonce_base64 = fido2.challenge.get_fido2_nonce()
            challenge_validity: int = int(get_from_config(FIDO2ConfigOptions.CHALLENGE_VALIDITY_TIME,
                                                          get_from_config('DefaultChallengeValidityTime', 120)))
            challenge: Challenge = Challenge(serial=self.token.serial,
                                             transaction_id=None,  # will be generated by the challenge
                                             challenge=nonce_base64,
                                             data=None,
                                             session="",
                                             validitytime=challenge_validity)
            challenge.save()

            # User ID
            fido2_user_id = base64url_to_bytes(
                token_user.attributes[
                    FIDO2TokenInfo.USER_ID]) if FIDO2TokenInfo.USER_ID in token_user.attributes else None

            # Excluded Credentials
            reg_ids = get_optional(params, "registered_credential_ids") or []
            registered_credential_ids: list[bytes] = [base64url_to_bytes(cred_id) for cred_id in reg_ids]
            excluded_credentials: list[PublicKeyCredentialDescriptor] = ([PublicKeyCredentialDescriptor(id=cred)
                                                                          for cred in registered_credential_ids])

            # Key Algorithms
            pub_key_algorithms = get_optional(params, FIDO2PolicyAction.PUBLIC_KEY_CREDENTIAL_ALGORITHMS,
                                              default=[COSEAlgorithmIdentifier.ECDSA_SHA_256,
                                                       COSEAlgorithmIdentifier.RSASSA_PKCS1_v1_5_SHA_256])

            # User Verification
            user_verification: UserVerificationRequirement = UserVerificationRequirement.PREFERRED
            if FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT in params:
                user_verification = UserVerificationRequirement(params[FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT])

            # Attestation (None is recommended for passkeys)
            attestation: AttestationConveyancePreference = AttestationConveyancePreference.NONE
            if PasskeyAction.AttestationConveyancePreference in params:
                attestation = AttestationConveyancePreference(params[PasskeyAction.AttestationConveyancePreference])

            registration_options: PublicKeyCredentialCreationOptions = generate_registration_options(
                rp_id=rp_id,
                rp_name=rp_name,
                user_name=token_user.login,
                user_display_name=token_user.login,
                user_id=fido2_user_id,
                attestation=attestation,
                authenticator_selection=AuthenticatorSelectionCriteria(
                    resident_key=ResidentKeyRequirement.REQUIRED,
                    user_verification=user_verification,
                ),
                challenge=base64url_to_bytes(nonce_base64),
                exclude_credentials=excluded_credentials,
                supported_pub_key_algs=pub_key_algorithms,
                timeout=12000,
            )

            # Save the userid if there was none before
            if not fido2_user_id:
                fido2_user_id = registration_options.user.id
                token_user.set_attribute(FIDO2TokenInfo.USER_ID, bytes_to_base64url(fido2_user_id))

            options_json: str = options_to_json(registration_options)
            response_detail["passkey_registration"] = json.loads(options_json)
            response_detail["transaction_id"] = challenge.transaction_id

            # Add RP ID, Name and user_id to the token info
            self.add_tokeninfo_dict({
                FIDO2TokenInfo.RELYING_PARTY_ID: rp_id,
                FIDO2TokenInfo.RELYING_PARTY_NAME: rp_name,
                FIDO2TokenInfo.USER_ID: bytes_to_base64url(fido2_user_id)
            })
        else:
            response_detail = {}
        return response_detail

    def update(self, param, reset_failcount=True):
        """
        Second step of enrollment: Verify the registration data from the authenticator with the challenge from the
        database. If the registration is successful, the token is set to enrolled and metadata is written to the token
        info.
        To complete the registration, the following parameters are required in param:
        - attestationObject
        - clientDataJSON
        - credential_id
        - rawId
        - authenticatorAttachment
        - transaction_id
        - HTTP_ORIGIN
        - FIDO2PolicyAction.RELYING_PARTY_ID ("webauthn_relying_party_id")
        """
        response_detail = {"details": {"serial": self.token.serial}}

        attestation = get_optional(param, "attestationObject")
        client_data = get_optional(param, "clientDataJSON")

        if not (attestation and client_data) and not self.token.rollout_state == ROLLOUTSTATE.CLIENTWAIT:
            self.token.rollout_state = ROLLOUTSTATE.CLIENTWAIT
            self.token.active = False
            # Set the description in the first enrollment step
            if "description" in param:
                self.set_description(param["description"])

        elif attestation and client_data and self.token.rollout_state == ROLLOUTSTATE.CLIENTWAIT:
            # Finalize the registration by verifying the registration data from the authenticator
            credential_id = get_required(param, "credential_id")
            credential_id_raw = get_required(param, "rawId")
            authenticator_attachment = get_required(param, "authenticatorAttachment")
            transaction_id = get_required(param, "transaction_id")
            expected_rp_id = get_required(param, FIDO2PolicyAction.RELYING_PARTY_ID)
            expected_origin = get_required(param, "HTTP_ORIGIN")

            serial = self.token.serial
            challenges = [challenge for challenge in get_challenges(serial=serial, transaction_id=transaction_id)
                          if challenge.is_valid()]

            if not len(challenges):
                raise EnrollmentError(f"The enrollment challenge does not exist or has timed out for {serial}")
            try:
                registration_verification: VerifiedRegistration = verify_registration_response(
                    credential={
                        "id": credential_id,
                        "rawId": credential_id_raw,
                        "response": {
                            "attestationObject": attestation,
                            "clientDataJSON": client_data
                        },
                        "type": "public-key",
                        "authenticatorAttachment": authenticator_attachment,
                    },
                    expected_challenge=challenges[0].challenge.encode("utf-8"),
                    expected_origin=expected_origin,
                    expected_rp_id=expected_rp_id,
                )
            except InvalidRegistrationResponse as ex:
                log.error(f"Invalid registration response: {ex}")
                raise EnrollmentError(f"Invalid registration response: {ex}")
            except InvalidJSONStructure as ex:
                log.error(f"Invalid JSON structure: {ex}")
                raise EnrollmentError(f"Invalid JSON structure: {ex}")

            # Verification successful, set the token to enrolled and save information returned by the authenticator
            self.token.rollout_state = ROLLOUTSTATE.ENROLLED
            # Protect the credential_id by setting it as the token secret
            self.set_otpkey(bytes_to_base64url(registration_verification.credential_id))

            # Token Info
            credential_id_hash = hash_credential_id(credential_id)
            token_info: dict = {
                FIDO2TokenInfo.DEVICE_TYPE: registration_verification.credential_device_type,
                FIDO2TokenInfo.BACKED_UP: registration_verification.credential_backed_up,
                FIDO2TokenInfo.PUBLIC_KEY: bytes_to_base64url(registration_verification.credential_public_key),
                FIDO2TokenInfo.AAGUID: registration_verification.aaguid,
                FIDO2TokenInfo.SIGN_COUNT: registration_verification.sign_count,
                FIDO2TokenInfo.CREDENTIAL_ID_HASH: credential_id_hash
            }
            # Save the credential_id hash to an extra table to be able to find the token faster
            save_credential_id_hash(credential_id_hash, self.token.id)

            # If the attestation object contains a x5c certificate, save it in the token info
            # and set the description to the CN if it is not already set
            if registration_verification.attestation_object:
                attestation_object: AttestationObject = parse_attestation_object(
                    registration_verification.attestation_object)
                if attestation_object.att_stmt and attestation_object.att_stmt.x5c:
                    attestation_certificate: Certificate = cryptography.x509.load_der_x509_certificate(
                        attestation_object.att_stmt.x5c[0])
                    certificate_pem: str = (attestation_certificate.public_bytes(serialization.Encoding.PEM)
                                            .decode("utf-8").replace("\n", ""))
                    token_info[FIDO2TokenInfo.ATTESTATION_CERTIFICATE] = certificate_pem
                    if not self.token.description:
                        attributes: list[NameAttribute] = attestation_certificate.subject.get_attributes_for_oid(
                            NameOID.COMMON_NAME)
                        if attributes:
                            self.set_description(attributes[0].value)
            self.add_tokeninfo_dict(token_info)
            self.token.active = True
            # Remove the challenge
            challenges[0].delete()
        return response_detail

    @check_token_locked
    def check_otp(self, otpval, counter=None, window=None, options=None):
        """
        :param otpval: Unused for this token type
        :type otpval: None
        :param counter: Unused for this token type
        :type counter: int
        :param window: Unused for this token type
        :type window: None
        :param options: Contains the data from the client, along with policy configurations.
                        For compatibility with the WebAuthnTokenClass, some keys can have multiple names.
                        The following keys are required:
                        - "challenge"
                        - "authenticatorData" or "authenticatordata"
                        - "clientDataJSON" or "clientdata"
                        - "signature" or "signaturedata"
                        - "userHandle" or "userhandle"
                        - "HTTP_ORIGIN"
                        The following keys are optional:
                        - "webauthn_user_verification_requirement", defaults to preferred

        :type options: dict
        :return: A numerical value where values larger than zero indicate success.
        :rtype: int
        """
        authenticator_data = get_required_one_of(options, ["authenticatorData", "authenticatordata"])
        client_data_json = get_required_one_of(options, ["clientDataJSON", "clientdata"])
        signature = get_required_one_of(options, ["signature", "signaturedata"])
        user_handle = get_optional_one_of(options, ["userHandle", "userhandle"])
        expected_challenge = get_required(options, "challenge").encode("utf-8")
        expected_origin = get_required(options, "HTTP_ORIGIN")
        user_verification = get_optional(options, "user_verification", "preferred")

        credential_id = self.token.get_otpkey().getKey().decode("utf-8")
        rp_id = self.get_tokeninfo(FIDO2TokenInfo.RELYING_PARTY_ID)

        try:
            verified_authentication: VerifiedAuthentication = verify_authentication_response(
                credential={
                    "id": credential_id,
                    "rawId": credential_id,
                    "response": {
                        "authenticatorData": authenticator_data,
                        "clientDataJSON": client_data_json,
                        "signature": signature,
                        "userHandle": user_handle,
                    },
                    "type": "public-key",
                    "authenticatorAttachment": "cross-platform",
                    "clientExtensionResults": {},
                },
                expected_challenge=expected_challenge,
                expected_rp_id=rp_id,
                expected_origin=expected_origin,
                require_user_verification=user_verification == "required",
                credential_current_sign_count=int(self.get_tokeninfo("sign_count")),
                credential_public_key=base64url_to_bytes(self.get_tokeninfo("public_key")),
            )
        except InvalidAuthenticationResponse as ex:
            log.info(f"Passkey authentication failed: {ex}")
            return -1

        self.add_tokeninfo("sign_count", verified_authentication.new_sign_count)
        return 1

    @classmethod
    def get_default_challenge_text_auth(cls) -> str:
        return str(lazy_gettext("Please authenticate with your passkey!"))

    @classmethod
    def get_default_challenge_text_register(cls) -> str:
        return str(lazy_gettext("Please confirm the registration with your passkey!"))

    def create_challenge(self, transactionid=None, options=None):
        """
        Passkey does not create a challenge itself, it uses an open challenge acquired from /validate/initialize.
        By returning False here, passkey tokens will not generate a challenge via
        /validate/triggerchallenge -> create_challenge_from_tokens()
        Optionally, creating a challenge can be enabled by setting the passkey_trigger_by_pin policy
        """
        if options and PasskeyAction.EnableTriggerByPIN in options and options[PasskeyAction.EnableTriggerByPIN]:
            rp_id = get_required(options, FIDO2PolicyAction.RELYING_PARTY_ID)
            user_verification = get_optional(options, "user_verification", "preferred")
            challenge = fido2.challenge.create_fido2_challenge(rp_id, user_verification=user_verification,
                                                               transaction_id=transactionid, serial=self.token.serial)
            message = challenge["message"]
            transaction_id = challenge["transaction_id"]
            challenge_details = {"challenge": challenge["challenge"], "rpId": rp_id,
                                 "userVerification": user_verification}
            # TODO this vvv is horrible
            return True, message, transaction_id, challenge_details
        else:
            return False, "", "", {}

    @log_with(log)
    def use_for_authentication(self, options):
        return self.is_active()

    def inc_failcount(self):
        """
        Do not increment the fail count for passkey, since their authentication process is decoupled from the usual.
        """
        pass

    @classmethod
    def is_multichallenge_enrollable(cls):
        return True

    @log_with(log)
    def is_challenge_request(self, passw, user=None, options=None):
        """
        This token type is always challenge-response. If the pin matches, a challenge should be created.
        """
        if options and PasskeyAction.EnableTriggerByPIN in options and options[PasskeyAction.EnableTriggerByPIN]:
            return self.check_pin(passw, user=user, options=options)
        return False

    @check_token_locked
    def authenticate(self, passw, user=None, options=None):
        """
        This is called from check_tokenlist. Suppress missing params here so "wrong otp value" is returned if no
        authentication could be made.
        """
        try:
            authenticator_data = get_required_one_of(options, ["authenticatorData", "authenticatordata"])
            client_data_json = get_required_one_of(options, ["clientDataJSON", "clientdata"])
            signature = get_required_one_of(options, ["signature", "signaturedata"])
            user_handle = get_optional_one_of(options, ["userHandle", "userhandle"])
            expected_challenge = get_required(options, "challenge").encode("utf-8")
            expected_origin = get_required(options, "HTTP_ORIGIN")
            user_verification = get_optional(options, "user_verification", "preferred")
        except ParameterError as e:
            log.debug(f"Missing parameter for authentication with passkey: {e}")
            # TODO authenticate has horrible return values
            return False, -1, None

        pin_match = self.check_pin(passw, user=user, options=options)
        if not pin_match:
            return False, -1, None
        otp_match = self.check_otp(None, 0, None, options)
        return pin_match, otp_match, None



"""
This test file tests the LDAP machine resolver in
lib/machines/ldap.py
"""

HOSTSFILE = "tests/testdata/hosts"
from .base import MyTestCase
from privacyidea.lib.machines.ldap import LdapMachineResolver
from privacyidea.lib.machines.base import MachineResolverError
from . import ldap3mock
import netaddr

LDAPDirectory = [{"dn": "cn=admin,ou=example,o=test",
                  "attributes": {"cn": "admin",
                                 "userPassword": "secret"}
                  },
                 {"dn": "cn=machine1,ou=example,o=test",
                  "attributes": {'cn': 'machine1',
                                 "objectClass": "computer",
                                 "dNSHostName": "machine1.example.test",
                                 "iPAddress": "1.2.3.4"}
                 },
                 {"dn": 'cn=machine2,ou=example,o=test',
                  "attributes": {'cn': 'machine2',
                                 "objectClass": "computer",
                                 "dNSHostName": "machine2.example.test"}
                 },
                 {"dn": 'cn=machine3,ou=example,o=test',
                  "attributes": {'cn': 'machine3',
                                 "objectClass": "computer",
                                 "dNSHostName": "machine3.example.test",
                                }
                 }]

MYCONFIG = {"HOSTNAMEATTRIBUTE": "dNSHostName",
            "LDAPURI": "ldap://1.2.3.4",
            "IDATTRIBUTE": "DN",
            "LDAPBASE": "o=test",
            "BINDDN": "cn=admin,ou=example,o=test",
            "BINDPW": "secret",
            "SEARCHFILTER": "(objectClass=computer)",
            "REVERSEFITLER": "(&(dNSHostName=%s)(objectClass=computer))"}


class LdapMachineTestCase(MyTestCase):
    """
    Test the LDAP Resolver
    """
    mreso = LdapMachineResolver("myResolver",
                                config=MYCONFIG)

    mrAD = LdapMachineResolver("myreso",
                               config={"LDAPURI": "ldap://172.16.200.202",
                                       "LDAPBASE": "dc=privacyidea,"
                                                   "dc=test",
                                       "BINDDN":
                                           "administrator@privacyidea.test",
                                       "BINDPW": "Test1234!",
                                       "HOSTNAMEATTRIBUTE": "dNSHostName"})


    def test_01_get_config_description(self):
        # Missing LDAPURI
        self.assertRaises(MachineResolverError, LdapMachineResolver,
                          "reso2", config={"IDATTRIBUTE": "na"})

        # Missing LDAPBASE
        self.assertRaises(MachineResolverError, LdapMachineResolver,
                          "reso2", config={"LDAPURI": "ldap://1.2.3.4"})

        desc = self.mreso.get_config_description()
        self.assertEqual(desc.get("ldap").get("config").get("LDAPURI"),
                         "string")

    """
    def test_01_get_machines(self):
        machines = self.mrAD.get_machines()
        self.assertEqual(len(machines), 1)
        self.assertEqual(machines[0].hostname, "dc01.privacyidea.test")

    def test_02_get_machine_id(self):
        id = self.mrAD.get_machine_id(hostname="dc01.privacyidea.test")
        self.assertEqual(id,
                         "CN=DC01,OU=Domain Controllers,DC=privacyidea,DC=test")
    """

    @ldap3mock.activate
    def test_03_get_machines(self):
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        machines = self.mreso.get_machines()
        self.assertEqual(len(machines), 3)

        machines = self.mreso.get_machines(hostname="machine1.example.test")
        self.assertEqual(len(machines), 1)

        # THere is one machine, that contains "e2"
        machines = self.mreso.get_machines(hostname="e2",
                                           substring=True)
        self.assertEqual(len(machines), 1)

    def test_04_get_machines_any(self):
        # The MockLdap module can not handle complicates ldap searches like
        # (&(&( )( )( ))(|( )( )( )) )
        # so we only test the creation of the ldapsearch string
        filter = LdapMachineResolver.\
            _create_ldap_filter("(""objectClass=computer)",
                                "objectSid", "",
                                "dNSHostName", "",
                                "", "",
                                substring=True, any="substr")
        self.assertEqual(filter, "(&(&(objectClass=computer))"
                                 "(|(objectSid=*substr*)("
                                 "dNSHostName=*substr*)))")



    def test_05_get_single_machine(self):

        machines = self.mreso.get_machines(machine_id="cn=machine1,ou=example,o=test")
        self.assertEqual(len(machines), 1)
        machine = machines[0]
        self.assertEqual(machine.id, "cn=machine1,ou=example,o=test")
        self.assertTrue(machine.has_hostname("machine1.example.test"))

    def test_06_get_machine_id(self):
        id = self.mreso.get_machine_id(hostname="not existing")
        self.assertTrue(id is None)

        id = self.mreso.get_machine_id(hostname="machine1.example.test")
        self.assertEqual(id, "cn=machine1,ou=example,o=test")

    @ldap3mock.activate
    def test_07_testconnection(self):
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        (success, desc) = LdapMachineResolver.testconnection(MYCONFIG)
        self.assertTrue(success)
        self.assertTrue("3" in desc)
        # "Your LDAP config seems to be OK, 3 machine objects found."

    @ldap3mock.activate
    def test_08_start_tls(self):
        # Check that START_TLS and TLS_VERIFY are actually passed to the ldap3 Connection
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        config = MYCONFIG.copy()
        config['START_TLS'] = '1'
        config['TLS_VERIFY'] = '1'
        start_tls_resolver = LdapMachineResolver("myResolver", config=config)
        machines = start_tls_resolver.get_machines()
        self.assertEqual(len(machines), 3)
        # We check two things:
        # 1) start_tls has actually been called!
        self.assertTrue(start_tls_resolver.l.start_tls_called)
        # 2) All Server objects were constructed with a non-None TLS context, but use_ssl=False
        for _, kwargs in ldap3mock.get_server_mock().call_args_list:
            self.assertIsNotNone(kwargs['tls'])
            self.assertFalse(kwargs['use_ssl'])


    @ldap3mock.activate
    def test_09_ldaps(self):
        # Check that use_ssl and tls are actually passed to the Connection
        ldap3mock.setLDAPDirectory(LDAPDirectory)
        config = MYCONFIG.copy()
        config['LDAPURI'] = 'ldaps://1.2.3.4'
        config['TLS_VERIFY'] = '1'
        ldaps_resolver = LdapMachineResolver("myResolver", config=config)
        machines = ldaps_resolver.get_machines()
        self.assertEqual(len(machines), 3)
        # We check that all Server objects were constructed with a non-None TLS context and use_ssl=True
        for _, kwargs in ldap3mock.get_server_mock().call_args_list:
            self.assertIsNotNone(kwargs['tls'])
            self.assertTrue(kwargs['use_ssl'])


#  2018-11-22 Initial create
#             Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This is the BaseClass for monitoring statistics data.

The monitoring database is a time series database. It is supposed to store
tuples of (timestamp, status key, status value)

A monitoring module needs to provide the possibility to write new data,
return all available keys, return the last data. 
"""
import logging
log = logging.getLogger(__name__)
from privacyidea.lib.log import log_with


class Monitoring(object):

    def __init__(self, config=None):
        pass

    def get_keys(self):
        """
        Return a list of the available statistic keys.

        :return: list of identifiers
        """
        return []

    def get_values(self, stats_key, start_timestamp=None, end_timestamp=None):
        """
        Return a list of tuples of (timestamp, value) for the requested stats_key.

        :param stats_key: Identifier of the stats
        :param start_timestamp: start of the time frame
        :type start_timestamp: timezone aware datetime
        :param end_timestamp:  end of the time frame
        :type end_timestamp: timezone aware datetime
        :return:
        """
        return []

    def get_last_value(self, stats_key):
        """
        returns the last value of the given stats_key in time.
        :param stats_key: The identifier of the stats
        :return: a string value.
        """
        pass

    def delete(self, stats_key, start_timestamp, end_timestamp):
        """
        Delete all entries of the stats_key for the given time frame.
        The start_timestamp and end_timestamp are also deleted.

        :param stats_key: Identifier of the stats
        :param start_timestamp: beginning of the time frame
        :type start_timestamp: timezone aware datetime
        :param end_timestamp: end of the time frame
        :type end_timestamp: timezone aware datetime
        :return: number of deleted entries
        """
        return 0

    def add_value(self, stats_key, stats_value, timestamp, reset_values=False):
        """
        This method adds a measurement point to the statistics key "stats_key".
        If reset_values is set to True, all old values of this stats_key are deleted.

        :param stats_key: Identifier of the stats
        :param stats_value: measured value
        :param timestamp: the timestamp of the measurement
        :type timestamp: timezone aware datetime
        :param reset_values: boolean to indicate the reset
        :return: None
        """
        pass







class FIDO2PolicyAction:
    ALLOWED_TRANSPORTS = 'webauthn_allowed_transports'
    TIMEOUT = 'webauthn_timeout'
    RELYING_PARTY_NAME = 'webauthn_relying_party_name'
    RELYING_PARTY_ID = 'webauthn_relying_party_id'
    AUTHENTICATOR_ATTACHMENT = 'webauthn_authenticator_attachment'
    AUTHENTICATOR_SELECTION_LIST = 'webauthn_authenticator_selection_list'
    USER_VERIFICATION_REQUIREMENT = 'webauthn_user_verification_requirement'
    PUBLIC_KEY_CREDENTIAL_ALGORITHMS = 'webauthn_public_key_credential_algorithms'
    AUTHENTICATOR_ATTESTATION_FORM = 'webauthn_authenticator_attestation_form'
    AUTHENTICATOR_ATTESTATION_LEVEL = 'webauthn_authenticator_attestation_level'
    REQ = 'webauthn_req'
    AVOID_DOUBLE_REGISTRATION = 'webauthn_avoid_double_registration'


class PasskeyAction:
    AttestationConveyancePreference = "passkey_attestation_conveyance_preference"
    EnableTriggerByPIN= "passkey_trigger_by_pin"


#  2015-02-27 Cornelius Kölbel <cornelius@privacyidea.org>
#             Initial writup
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
"""
This is the library for retrieving machines and adding tokens and
applications to machines.

This module is tested in tests/test_lib_machinetokens.py and
tests/test_lib_machines.py.

It depends on the database model models.py and on the machineresolver
lib/machineresolver.py, so this can be tested standalone without realms,
tokens and webservice!
"""
from .error import ResourceNotFoundError, ParameterError
from .machineresolver import get_resolver_list, get_resolver_object
from privacyidea.models import Token
from privacyidea.models import (MachineToken, db, MachineTokenOptions,
                                MachineResolver, get_token_id,
                                get_machineresolver_id,
                                get_machinetoken_ids)
from privacyidea.lib.utils import fetch_one_resource
from netaddr import IPAddress
from sqlalchemy import and_
import logging
import re

from .token import token_exist

log = logging.getLogger(__name__)
from privacyidea.lib.log import log_with
from privacyidea.lib.applications.base import get_auth_item, get_machine_application_class_dict

ANY_MACHINE = "any machine"
NO_RESOLVER = "no resolver"


@log_with(log)
def get_machines(hostname=None, ip=None, id=None, resolver=None, any=None,
                 substring=True):
    """
    This returns a list of machines from ALL resolvers matching this criterion.

    :param hostname: The hostname of the machine, substring matching
    :type hostname: basestring
    :param ip: The IPAddress of the machine
    :type ip: netaddr.IPAddress
    :param id: The id of the machine, substring matching
    :type id: basestring
    :param resolver: The resolver of the machine, substring matching
    :type resolver: basestring
    :param any: a substring, that matches EITHER of hostname, ip or resolver
    :type any: basestring
    :param substring: If True, machines are search with the parameters as
        substrings
    :return: list of Machine Objects.
    """
    resolver_list = get_resolver_list()
    all_machines = []

    for reso in resolver_list.keys():
        # The resolvernames are the keys of the dictionary
        if resolver and resolver not in reso:
            # filter for other resolvers
            continue
        reso_obj = get_resolver_object(reso)
        resolver_machines = reso_obj.get_machines(hostname=hostname,
                                                  ip=ip,
                                                  machine_id=id,
                                                  any=any,
                                                  substring=substring)
        all_machines += resolver_machines

    return all_machines


def get_hostname(ip):
    """
    Return a hostname for a given IP address
    :param ip: IP address
    :type ip: IPAddress
    :return: hostname
    """
    machines = get_machines(ip=ip)
    if len(machines) > 1:
        raise Exception("Can not get unique ID for IP=%r. "
                        "More than one machine found." % ip)
    if len(machines) == 1:
        # There is only one machine in the list and we get its ID
        hostname = machines[0].hostname
        # return the first hostname
        if type(hostname) == list:
            hostname = hostname[0]
    else:
        raise Exception("There is no machine with IP={0!r}".format(ip))
    return hostname


def get_machine_id(hostname, ip=None):
    """
    Determine the id for a given hostname.
    The ID consists of the resolvername and the machine_id in this resolver.

    :param hostname: The hostname of the machine
    :param ip: The optional IP of the machine
    :return: tuple of machine_id and resolvername
    """
    machine_id = None
    resolver_name = None
    machines = get_machines(hostname=hostname, ip=ip, substring=False)
    if len(machines) > 1:
        raise Exception("Can not get unique ID for hostname=%r and IP=%r. "
                        "More than one machine found." % (hostname, ip))
    if len(machines) == 1:
        # There is only one machine in the list and we get its ID
        machine_id = machines[0].id
        resolver_name = machines[0].resolver_name

    if machine_id is None:
        raise Exception("There is no machine with name={0!r} and IP={1!r}".format(hostname, ip))

    return machine_id, resolver_name


#
#  Attach tokens to machines
#


@log_with(log)
def attach_token(serial, application, hostname=None, machine_id=None, resolver_name=None, options=None):
    """
    Attach a token with an application to a machine. You need to provide
    either the hostname or the (machine_id, resolver_name) of the machine,
    to which you want to attach the token.

    :param serial: The serial number of the token
    :type serial: string
    :param application: The name of the application - something like ssh or luks
    :type application: basestring
    :param hostname: The hostname of the machine, to which your want to
    attach the token. If the hostname is not unique, an exception is raised.
    :type hostname: basestring
    :param machine_id: The machine ID of the machine, you want to attach the token to.
    :type machine_id: basestring
    :param resolver_name: The resolver_name of the machine you want to attach the token to.
    :type resolver_name: basestring
    :param options: additional options
    :return: the new MachineToken Object
    """
    # Check for valid serial
    if not token_exist(serial):
        raise ResourceNotFoundError(f"Token with serial {serial} does not exist.")

    # Check for valid application
    valid_applications = get_machine_application_class_dict().keys()
    if application not in valid_applications:
        raise ParameterError(
            f"Application {application} is not supported. Valid applications are {valid_applications}.")

    if hostname or machine_id or resolver_name:
        machine_id, resolver_name = _get_host_identifier(hostname, machine_id, resolver_name)
    # Now we have all data to create the MachineToken
    machine_token = MachineToken(machineresolver=resolver_name, machine_id=machine_id, serial=serial,
                                 application=application)
    machine_token.save()
    # Add options to the machine token
    if options:
        add_option(machine_token_id=machine_token.id, options=options)

    return machine_token


@log_with(log)
def detach_token(serial, application, hostname=None, machine_id=None, resolver_name=None, machine_token_id=None,
                 filter_params=None):
    """
    Delete a machine token.
    Also deletes the corresponding MachineTokenOptions
    You need to provide either the hostname or the (machine_id,
    resolver_name) of the machine, from which you want to detach the
    token.

    :param serial: The serial number of the token
    :type serial: string
    :param application: The name of the application - something like ssh or luks
    :type application: basestring
    :param hostname: The hostname of the machine
    If the hostname is not unique, an exception is raised.
    :type hostname: basestring
    :param machine_id: The machine ID of the machine
    :type machine_id: basestring
    :param resolver_name: The resolver_name of the machine
    :type resolver_name: basestring
    :param machine_token_id: The specific id of a specific machine token
    :param filter_params: additional machine token options to filter the detachment
    :type filter_params: dict
    :return: the number of deleted MachineToken objects
    """
    r = None
    if machine_token_id:
        # We have specific machine_token_id, which we are supposed to delete.
        MachineTokenOptions.query.filter(MachineTokenOptions.machinetoken_id == int(machine_token_id)).delete()
        # Delete MachineToken
        r = MachineToken.query.filter(MachineToken.id == int(machine_token_id)).delete()
    else:
        filter_params = filter_params or {}
        if machine_id == ANY_MACHINE and resolver_name == NO_RESOLVER:
            # For later handling we need to reset the machine_id and the resolver_name:
            machine_id = None
            resolver_name = None

        machine_tokens = list_token_machines(serial)
        # Delete MachineTokenOptions
        for machine_token in machine_tokens:
            if (machine_token.get("application") == application
                    and ((not machine_id or machine_id == machine_token.get("machine_id")
                          and (not hostname or hostname == machine_token.get("hostname"))
                          and (not resolver_name or resolver_name == machine_token.get("resolver"))))):
                delete_mt = True
                for key, value in filter_params.items():
                    # Check if the machine token contains the correct filter values
                    if machine_token.get("options").get(key) != value:
                        delete_mt = False
                if delete_mt:
                    MachineTokenOptions.query.filter(
                        MachineTokenOptions.machinetoken_id == machine_token.get("id")).delete()
                    # Delete MachineToken
                    r = MachineToken.query.filter(MachineToken.id == machine_token.get("id")).delete()
    db.session.commit()
    return r


def add_option(machine_token_id=None, machine_id=None, resolver_name=None,
               hostname=None, serial=None, application=None, options=None):
    """
    Add options to the machine token definition.
    You can either specify machine_token_id or
     * machine_id, resolver_name, serial, application or
     * hostname, serial, application
    :param machine_token_id: The database id of the machine token
    :param machine_id: the resolver dependent machine id
    :param resolver_name: the name of the machine resolver
    :param hostname: the machine name
    :param serial: the serial number of the token
    :param application: the application
    :param options: options to add
    """
    if options is None:
        options = {}
    if machine_token_id:
        machine_token_ids = [machine_token_id]
    else:
        machine_id, resolver_name = _get_host_identifier(hostname, machine_id, resolver_name)

        machine_token_ids = get_machinetoken_ids(machine_id, resolver_name, serial, application)

    for option_name, option_value in options.items():
        for machine_token_id in machine_token_ids:
            MachineTokenOptions(machine_token_id, option_name, option_value)
    return len(options)


def delete_option(machine_token_id=None, machine_id=None, resolver_name=None,
                  hostname=None, serial=None, application=None, key=None):
    """
    delete option from a machine token definition

    You can either specify machine_token_id or
     * machine_id, resolver_name, serial, application or
     * hostname, serial, application
    :param machine_token_id: The database id of the machine token
    :param machine_id: the resolver dependent machine id
    :param resolver_name: the name of the machine resolver
    :param hostname: the machine name
    :param serial: the serial number of the token
    :param application: the application
    :param key: the key of the option
    """
    if machine_token_id:
        machine_token_ids = [int(machine_token_id)]
    else:
        machine_id, resolver_name = _get_host_identifier(hostname, machine_id, resolver_name)
        machine_token_ids = get_machinetoken_ids(machine_id, resolver_name, serial, application)

    res = 0
    for machine_token_id in machine_token_ids:
        res = MachineTokenOptions.query.filter(and_(
            MachineTokenOptions.machinetoken_id == machine_token_id,
            MachineTokenOptions.mt_key == key)).delete()
    db.session.commit()
    return res


@log_with(log)
def list_machine_tokens(hostname=None, machine_id=None, resolver_name=None, serial=None, application=None,
                        filter_params=None, serial_pattern=None):
    """
    Returns a list of tokens assigned to the given machine.

    :return: JSON of all tokens connected to machines with the corresponding
             application.
    """
    res = []
    filter_params = filter_params or {}
    service_id = filter_params.get("service_id")
    # The service_id overrules the hostname
    if not service_id and (hostname or machine_id or resolver_name):
        machine_id, resolver_name = _get_host_identifier(hostname, machine_id, resolver_name)
        machine_resolver_id = get_machineresolver_id(resolver_name)
        sql_query = MachineToken.query.filter(and_(MachineToken.machine_id == machine_id,
                                                   MachineToken.machineresolver_id == machine_resolver_id))
    else:
        # If we have no specific machine defined, we find all applications/serials
        sql_query = MachineToken.query.filter()

    if application:
        sql_query = sql_query.filter(MachineToken.application == application)
    if serial:
        # discrete serial
        token_id = get_token_id(serial)
        sql_query = sql_query.filter(MachineToken.token_id == token_id)

    for row in sql_query.all():
        # row.token contains the database token
        option_list = row.option_list
        options = {}
        for option in option_list:
            options[option.mt_key] = option.mt_value
        include_machine_token = True
        # check serial_pattern
        if serial_pattern:
            if not re.match(serial_pattern, row.token.serial, re.I):
                include_machine_token = False
        # we still think, it should be included
        if include_machine_token:
            for key, value in filter_params.items():
                token_option_value = options.get(key, "")
                if "*" in value:
                    # Simple wildcard matching
                    pattern = value.replace("*", ".*")
                    # We do a case-insensitive match
                    if not re.match(pattern, token_option_value, re.I):
                        include_machine_token = False
                elif token_option_value != value:
                    include_machine_token = False
        if include_machine_token:
            res.append({"serial": row.token.serial,
                        "machine_id": machine_id,
                        "resolver": resolver_name,
                        "type": row.token.tokentype,
                        "application": row.application,
                        "id": row.id,
                        "options": options})

    return res


@log_with(log)
def list_token_machines(serial):
    """
    This method returns the machines for a given token

    :return: returns a list of machines and apps
    """
    res = []
    db_token = fetch_one_resource(Token, serial=serial)

    for machine in db_token.machine_list:
        hostname = "unknown"
        if machine.machine_id and machine.machineresolver_id:
            machine_resolver = fetch_one_resource(MachineResolver, id=machine.machineresolver_id)
            resolver_name = machine_resolver.name
        else:
            resolver_name = NO_RESOLVER
            hostname = "any host"

        option_list = machine.option_list
        options = {}
        for option in option_list:
            options[option.mt_key] = option.mt_value

        # Try to determine the hostname
        machines = get_machines(id=machine.machine_id, resolver=resolver_name)
        if len(machines) == 1:
            # There is only one machine in the list and we get its ID
            hostname = machines[0].hostname
            # return the first hostname
            if type(hostname) == list:
                hostname = hostname[0]

        res.append({"machine_id": machine.machine_id or ANY_MACHINE,
                    "hostname": hostname,
                    "application": machine.application,
                    "resolver": resolver_name,
                    "options": options,
                    "id": machine.id,
                    "serial": serial})

    return res


def _get_host_identifier(hostname, machine_id, resolver_name):
    """
    This returns the combination of machine_id and resolver_name for some
    of the given values. This is used when attaching and detaching tokens to
    a machine to create a uniquely identifiable machine object.

    :param hostname:
    :param machine_id:
    :param resolver_name:
    :return:
    """
    if hostname:
        (machine_id, resolver_name) = get_machine_id(hostname)
    if not (machine_id and resolver_name):  # pragma: no cover
        raise Exception("Incomplete tuple of machine_id and resolver_name")

    return machine_id, resolver_name


def get_auth_items(hostname=None, application=None, serial=None, challenge=None, filter_param=None, user_agent=None):
    """
    Return the authentication items for a given hostname and the application.
    The hostname is used to identify the machine object. Then all attached
    tokens to this machine and its applications are searched.

    :param hostname:
    :param application:
    :param serial:
    :param challenge: A challenge for the auth_item
    :type challenge: basestring
    :param filter_param: Additional application specific parameter to filter
        the return value
    :type filter_param: dict
    :param user_agent: The user agent header of the request
    :type user_agent: str
    :return: dictionary of lists of the application auth items

    **Example response**:

    .. sourcecode:: json

       { "luks": [ { "slot": "....",
                     "partition": "....",
                     "challenge": "....",
                     "response": "...." }
                 ],
         "ssh": [ { "username": "....",
                    "sshkey": "...."},
                  { "username": "....",
                    "sshkey": "...." }
                 ] }
    """
    auth_items = {}
    machine_tokens = list_machine_tokens(hostname=hostname, serial=serial, application=application,
                                         filter_params=filter_param)

    for machine_token in machine_tokens:
        auth_item = get_auth_item(machine_token.get("application"),
                                  machine_token.get("type"),
                                  machine_token.get("serial"),
                                  challenge,
                                  options=machine_token.get("options"),
                                  filter_param=filter_param,
                                  user_agent=user_agent)
        if auth_item:
            if machine_token.get("application") not in auth_items:
                # we create a new empty list for the new application type
                auth_items[machine_token.get("application")] = []

            # Add the options to auth_item
            auth_item.update(machine_token.get("options", {}))
            auth_item["serial"] = serial

            # append the auth_item to the list
            auth_items[machine_token.get("application")].append(auth_item)

    return auth_items


def is_offline_token(serial: str) -> bool:
    """
    Verifies if the token with the given serial number is an offline token.

    :param serial: The serial number of the token
    :return: True if the token is an offline token, False otherwise
    """
    machines = list_token_machines(serial)
    machine_applications = [machine.get("application") for machine in machines]
    is_offline = "offline" in machine_applications
    return is_offline
